import json
import time
import jwt  # PyJWT library
import requests
from objict import objict
from rest import log
from rest import settings


DEBUG_FCM = settings.get("DEBUG_FCM", False)
logger = log.getLogger("fcm", filename="fcm.log")


def create_jwt(service_account_info):
    """
    Create a JWT (JSON Web Token) for service account authentication.
    
    :param service_account_info: Dictionary containing service account credentials.
    :return: JWT string.
    """
    now = int(time.time())
    payload = {
        'iss': service_account_info['client_email'],
        'sub': service_account_info['client_email'],
        'aud': 'https://oauth2.googleapis.com/token',
        'iat': now,
        'exp': now + 3600,
        'scope': 'https://www.googleapis.com/auth/firebase.messaging'
    }
    
    additional_headers = {
        'kid': service_account_info['private_key_id']
    }
    
    token = jwt.encode(payload, service_account_info['private_key'], algorithm='RS256', headers=additional_headers)
    return token

def get_access_token(jwt_token):
    """
    Exchange a JWT for an access token.
    
    :param jwt_token: JWT string.
    :return: Access token string.
    """
    url = 'https://oauth2.googleapis.com/token'
    headers = {
        'Content-Type': 'application/x-www-form-urlencoded'
    }
    payload = {
        'grant_type': 'urn:ietf:params:oauth:grant-type:jwt-bearer',
        'assertion': jwt_token
    }
    
    response = requests.post(url, headers=headers, data=payload)
    response_data = response.json()
    return response_data['access_token']


class FirebaseNotifier:
    def __init__(self, service_account_info):
        """
        Initialize the FirebaseNotifier with the project ID and service account info.
        
        :param project_id: The project ID from Firebase project settings.
        :param service_account_info: Dictionary containing service account credentials.
        """
        self.project_id = service_account_info.get("project_id")
        self.service_account_info = service_account_info
        self.fcm_url = f'https://fcm.googleapis.com/v1/projects/{self.project_id}/messages:send'
        self.jwt_token = None
        self._jwt_expires = 0
        self.access_token = None

    @property
    def is_jwt_expired(self):
        return time.time() > self._jwt_expires
    
    
    def _get_access_token(self):
        """
        Get the access token, refreshing if necessary.
        
        :return: Access token string.
        """
        if not self.jwt_token or self.is_jwt_expired:
            self.jwt_token = create_jwt(self.service_account_info)
            self._jwt_expires = time.time() + 3000
            self.access_token = get_access_token(self.jwt_token)
        return self.access_token
    
    def send(self, registration_token, title, body, data=None, **kwargs):
        """
        Send a notification to the specified registration token.
        
        :param registration_token: Device registration token.
        :param title: Title of the notification.
        :param body: Body of the notification.
        :param data: Optional data payload to send along with the notification.

        :param ttl: optional life of message in seconds 
        :return: Response from the FCM server.
        """
        access_token = self._get_access_token()
        headers = {
            'Content-Type': 'application/json',
            'Authorization': f'Bearer {access_token}',
        }
        message = dict(token=registration_token)
        if title and body:
            message["notification"] = dict(title=title, body=body)
        if data:
            priority = kwargs.get("priority", None)
            message["data"] = {k:str(v) for k,v in data.items()}
            # if priority == "high":
            #     message["android"] = dict(priority="high")
            #     message["apns"] = dict(headers={"apns-priority": "10"})
            # else:
            #     message["android"] = dict(priority="normal")
            #     message["apns"] = dict(headers={"apns-priority": "5"})
        payload = dict(message=message)
        if DEBUG_FCM:
            logger.info("request", payload)
        resp = requests.post(self.fcm_url, headers=headers, data=json.dumps(payload))
        if DEBUG_FCM:
            logger.info("response", resp.json())
        return resp



from . import models as auditlog
from . import cloudwatch

from rest import views as rv
from rest import decorators as rd
from datetime import datetime


@rd.url(r'^plog$')
@rd.url(r'^plog/(?P<plog_id>\d+)$')
@rd.perm_required('view_logs')
def plog_handler(request, plog_id=None):
    if not plog_id:
        min_pk = getattr(auditlog.settings, "PLOG_STALE_ID", 0)
        return auditlog.PersistentLog.on_rest_list(request, qset=auditlog.PersistentLog.objects.filter(pk__gte=min_pk))
    return auditlog.PersistentLog.on_rest_request(request, plog_id)


@rd.urlGET(r'^plog_old$')
@rd.perm_required('view_logs')
def plogList(request):
    auditlog.PersistentLog.on_request_handle()

    graph = request.DATA.get("graph", "default")
    qset = auditlog.PersistentLog.objects.all()
    if request.group:
        qset = qset.filter(group=request.group)

    ip = request.DATA.get("ip")
    if ip:
        qset = qset.filter(session__ip=ip)

    path = request.DATA.get("path")
    if path:
        qset = qset.filter(remote_path__icontains=path)
    
    method = request.DATA.get("method")
    if method:
        qset = qset.filter(method=method)

    action = request.DATA.get("action")
    if action:
        qset = qset.filter(action=action)
    
    component = request.DATA.get("component")
    if component:
        qset = qset.filter(component=component)
    
    pkey = request.DATA.get("pkey")
    if pkey:
        qset = qset.filter(pkey=pkey)

    username = request.DATA.get("username")
    if username:
        qset = qset.filter(user__username=username)
    
    term = request.DATA.get("term")
    if term:
        qset = qset.filter(message__icontains=term)
    
    return rv.restList(request, qset.order_by('-when'), **auditlog.PersistentLog.getGraph(graph))


@rd.url('server/logs')
@rd.perm_required('view_logs')
@rd.requires_params(["log_group", "log_streams"])
def cloudwatch_logs(request):
    log_group = request.DATA.get("log_group")
    log_streams = request.DATA.get("log_streams").split(',')
    pattern = request.DATA.get("search")
    if not pattern:
        data = dict(status=True, data=[], count=0)
        return rv.restResult(request, data)
    start = request.DATA.get("dr_start", "1d")
    end = request.DATA.get("dr_end", field_type=datetime)
    resp = cloudwatch.filter(log_group, log_streams, pattern,
                             start_time=start, end_time=end,
                             resp_format="nginx")
# else:
    #     resp = cloudwatch.get(log_group, log_streams, resp_format="nginx")
    lst = resp["events"]
    count = len(lst)
    data = dict(status=True, data=lst, count=len(lst), size=count)
    if "nextBackwardToken" in resp:
        data["backward"] = resp["nextBackwardToken"]
        data["forword"] = resp["nextForwardToken"]
    return rv.restResult(request, data)

@rd.urlPOST('cloudwatch/log')
@rd.perm_required('view_logs')
@rd.requires_params(["log_group", "log_stream", "event"])
def cloudwatch_log_groups(request):
    log_group = request.DATA.get("log_group")
    log_stream = request.DATA.get("log_stream")
    event = request.DATA.get("event")
    resp = cloudwatch.logToCloudWatch(event, log_group, log_stream)
    return rv.restResult(request, dict(status=True, data=resp))


@rd.urlGET('cloudwatch/log')
@rd.perm_required('cloudwatch')
@rd.requires_params(["log_group", "log_stream"])
def cloudwatch_log_groups(request):
    log_group = request.DATA.get("log_group")
    log_stream = request.DATA.get("log_stream")
    pattern = request.DATA.get("pattern")
    if pattern:
        start = request.DATA.get("dr_start", "6d")
        end = request.DATA.get("dr_end", field_type=datetime)
        resp = cloudwatch.filter(log_group, [log_stream], pattern,
                                 start_time=start, end_time=end,
                                 resp_format=request.DATA.get("format"))
    else:
        resp = cloudwatch.get(log_group, log_stream)
    lst = resp["events"]
    data = dict(status=True, data=lst, count=len(lst))
    if "nextBackwardToken" in resp:
        data["backward"] = resp["nextBackwardToken"]
        data["forword"] = resp["nextForwardToken"]
    data = dict(status=True, data=lst, count=len(lst))
    return rv.restResult(request, data)


@rd.urlGET('cloudwatch/groups')
@rd.perm_required('cloudwatch')
def cloudwatch_log_groups(request):
    return rv.restList(request, cloudwatch.getLogGroups(request.DATA.get("graph", None) is None))


@rd.urlPOST('cloudwatch/insights/start')
@rd.perm_required('cloudwatch')
@rd.requires_params(["log_groups"])
def cloudwatch_insights_start(request):
    log_groups = request.DATA.get("log_groups")
    log_streams = request.DATA.get("log_streams")
    start = request.DATA.get("dr_start", "6h")
    end = request.DATA.get("dr_end", field_type=datetime)
    ip = request.DATA.get("ip")
    text = request.DATA.get("text")
    query = request.DATA.get("query")
    if ip is None and text is None and query is None:
        return rv.restPermissionDenied(request)
    resp = cloudwatch.startSearch(log_groups, start, end, 
                                      ip=ip, text=text, query_string=query,
                                      log_streams=log_streams)
    return rv.restResult(request, dict(status=True, data=resp))


@rd.urlGET(r'cloudwatch/insights/results')
@rd.perm_required('cloudwatch')
@rd.requires_params(["query_id"])
def cloudwatch_insights_results(request):
    query_id = request.DATA.get("query_id")
    results = cloudwatch.getInsightResults(query_id, request.DATA.get("format"))
    return rv.restResult(request, results)




from .uberdict import UberDict
from binascii import hexlify
from collections import OrderedDict
from datetime import datetime

from io import StringIO
from threading import current_thread
import logging
import os
import errno
import sys
import threading
import time
import traceback
from .settings_helper import settings

LOG_FOLDER = settings.LOG_ROOT
if LOG_FOLDER is None:
    LOG_FOLDER = os.path.join(settings.BASE_DIR, "var", "logs")

COLOR_LOGS = True
MAX_LOG_SIZE = 10485760

# number of bytes to leave after rotating/truncating a file
ROTATE_LEFT_OVER_BYTES = 50000
LOG_COUNT = 3


class RestLogManager(object):
    def __init__(self):
        self.lock = threading.RLock()
        self.master = None
        self.loggers = {}
        self.streams = {}

    def setMaster(self, logger):
        self.acquire()
        self.master = logger
        self.release()

    def getStream(self, filename, max_bytes=MAX_LOG_SIZE):
        if filename is None:
            filename = "stdout"
        self.acquire()
        stream = self.streams.get(filename)
        if not stream:
            stream = RestLoggerStream(filename, max_bytes)
            self.streams[filename] = stream
        self.release()
        return stream

    def getLogger(self, name):
        self.loggers.get(name, None)

    def addLogger(self, logger):
        self.acquire()
        self.loggers[logger.name] = logger
        self.release()

    def removeLogger(self, logger):
        if logger.name in self.loggers:
            self.acquire()
            if logger.name in self.loggers:
                del self.loggers["name"]
            self.release()

    def acquire(self):
        """
        Acquire the I/O thread lock.
        """
        if self.lock:
            self.lock.acquire()

    def release(self):
        """
        Release the I/O thread lock.
        """
        if self.lock:
            self.lock.release()


LOG_MANAGER = RestLogManager()


def getLogger(
    name="root",
    filename=None,
    debug=False,
    errors_to_root=True,
    set_master=False,
    max_bytes=MAX_LOG_SIZE,
    create_path=True
):
    try:
        if name == "root":
            name == "rest"
        LOG_MANAGER.acquire()
        rest_logger = LOG_MANAGER.loggers.get(name, None)
        if rest_logger:
            if filename and not rest_logger.filename:
                rest_logger.setFilename(filename)
            LOG_MANAGER.release()
            return rest_logger
        level = RestLogger.INFO
        if debug:
            level = RestLogger.ALL
        #
        master = None
        if errors_to_root:
            master = LOG_MANAGER.master
        rest_logger = RestLogger(name, filename, level, master, max_bytes=max_bytes)
        LOG_MANAGER.loggers[name] = rest_logger
        if set_master:
            LOG_MANAGER.master = rest_logger
        if not os.path.exists(LOG_FOLDER):
            mkdir(LOG_FOLDER)
    except Exception as err:
        print((str(err)))
        print((str(traceback.format_exc())))
    LOG_MANAGER.release()
    return rest_logger


def setupMasterLogger(name="root", filename=None, debug=False):
    rest_logger = getLogger(name, filename, debug, False, set_master=True)
    rest_logger.capture_stdout()
    rest_logger.capture_stderr()
    return rest_logger


LEVEL_NAMES = {
    "50": "CRITICAL",
    "40": "ERROR",
    "30": "WARNING",
    "20": "INFO",
    "10": "DEBUG",
    "0": "ALL",
}

LEVEL_COLORS = {"WARNING": "YELLOW", "DEBUG": "BLUE", "CRITICAL": "PINK", "ERROR": "YELLOW"}


class RestLoggerStream(object):
    def __init__(self, filename, max_bytes=MAX_LOG_SIZE):
        self.is_file = False
        self.filename = filename
        if filename:
            if filename not in ["stdout", "stderr"]:
                self.is_file = True
                # lets put it in the var path
                if not os.path.exists(self.filename):
                    path, fname = os.path.split(filename)
                    if not len(path):
                        self.filename = os.path.join(LOG_FOLDER, filename)
        self.max_bytes = max_bytes
        self.stream = None
        self.is_stdout = False
        self.is_stderr = False
        self.lock = threading.RLock()
        self.open()

    def open(self):
        if self.stream:
            try:
                # test if file is still open?
                self.stream.size()
                return
            except:
                self.stream = None

        if self.is_file:
            try:
                self.acquire()
                self.stream = open(self.filename, "a")
            finally:
                self.release()
        else:
            self.stream = sys.stdout

    def close(self):
        if self.stream:
            try:
                self.acquire()
                self.stream.close()
                self.stream = None
            finally:
                self.release()

    def flush(self):
        if self.stream:
            self.stream.flush()

    def write(self, data):
        if not self.stream:
            self.open()
        self.stream.write(data)
        self.stream.flush()

    def acquire(self):
        """
        Acquire the I/O thread lock.
        """
        if self.lock:
            self.lock.acquire()

    def release(self):
        """
        Release the I/O thread lock.
        """
        if self.lock:
            self.lock.release()

    def size(self):
        if not self.is_file:
            return 0
        try:
            pos = self.stream.tell()
            self.stream.seek(0, 2)
            size = self.stream.tell()
            self.stream.seek(pos)
            return size
        except Exception:
            pass
        return 0

    def clear(self):
        self.acquire()
        if self.stream:
            self.stream.close()
            self.stream = None
        self.stream = open(self.filename, "w")
        self.release()

    def rotate(self):
        # assume we are already locked?
        # simple way lets open the file for writing
        # read the last
        # reopen file for reading and writing
        try:
            self.stream.close()
        except Exception:
            pass
        to_end = ""
        with open(self.filename, "r") as stream:
            try:
                stream.seek(0, 2)
                size = stream.tell()
                # now lets go back 3000 bytes and find first new line
                stream.seek(size - ROTATE_LEFT_OVER_BYTES)
                stream.readline()
                to_end = stream.read()
            except Exception:
                pass
        self.stream = open(self.filename, "w")
        self.stream.write(to_end)
        self.stream.flush()
        if self.is_stderr:
            self.capture_stderr()
        if self.is_stdout:
            self.capture_stdout()

    def rotateCheck(self):
        if self.is_file and self.max_bytes > 0:  # are we rolling over?
            self.open()
            # self.stream.seek(0, 2)  # due to non-posix-compliant Windows feature
            if self.size() >= self.max_bytes:
                return True
        return False

    def capture_stdout(self):
        self.is_stdout = True
        sys.stdout = self.stream

    def capture_stderr(self):
        self.is_stderr = True
        sys.stderr = self.stream


class RestLogger(object):
    CRITICAL = 50
    ERROR = 40
    WARNING = 30
    INFO = 20
    DEBUG = 10
    ALL = 0

    def __init__(
        self,
        name,
        filename=None,
        level=20,
        master=None,
        master_level=30,
        fmt=None,
        color=COLOR_LOGS,
        max_bytes=MAX_LOG_SIZE,
    ):
        self.name = name
        self.filename = filename
        self.level = level
        self.master = master
        self.master_level = master_level
        self.color = color
        if fmt is None:
            fmt = "{asctime} {levelname} {name}({threadName}): {msg}"
        self._fmt = fmt
        self.max_bytes = max_bytes
        self.lock = threading.RLock()
        self.stream = None
        self.setFilename(self.filename)

    def setFilename(self, filename):
        self.filename = filename
        try:
            self.stream = LOG_MANAGER.getStream(self.filename, self.max_bytes)
        except Exception as err:
            print("ERROR")
            print((str(err)))
            self.stream = LOG_MANAGER.getStream(None, self.max_bytes)
            print((str(err)))
            self.error(err)

    def setLevel(self, level):
        self.level = level

    def capture_stdout(self):
        sys.stdout = self.stream.stream

    def capture_stderr(self):
        sys.stderr = self.stream.stream

    def emit(self, record):
        if record.level >= self.level:
            record.name = self.name
            line = self.format(record)
            try:
                self.stream.acquire()
                if self.stream.rotateCheck():
                    # we need to rotate the logs
                    self.stream.rotate()
                self.stream.write(line)
            finally:
                self.stream.release()
        if self.master and record.level >= self.master_level:
            self.master.emit(record)

    def format(self, record):
        out = self._fmt.format(**record)
        level_color = None
        if self.color:
            level_color = LEVEL_COLORS.get(record.levelname, None)
            if level_color:
                out = "{}{}{}".format(ConsoleColors.__dict__[level_color], out, ConsoleColors.OFF)
        if record.stack:
            if level_color:
                out = "\n{}\n{}{}{}".format(out, ConsoleColors.RED, record.stack, ConsoleColors.OFF)
            else:
                out = "{}\n{}".format(out, record.stack)
        return out + "\n"

    def log(self, level, args, stack=None, master_only=False):
        if level < self.level and level < self.master_level:
            return
        record = UberDict()
        record.name = self.name
        record.asctime = datetime.now().strftime("%Y-%m-%d %H:%M:%S.%f")[:-3]
        record.level = level
        record.levelname = LEVEL_NAMES.get(str(level), str(level))
        record.threadName = current_thread().name
        record.stack = stack
        msg = []
        if type(args) in [list, tuple]:
            for arg in args:
                if isinstance(arg, dict) or isinstance(arg, list):
                    msg.append(prettyWriteToString(arg))
                else:
                    msg.append(str(arg))
            record.msg = "\n".join(msg)
        else:
            record.msg = args

        if master_only and self.master and level >= self.master_level:
            self.master.emit(record)
        else:
            self.emit(record)

    def insertLineBreak(self, count=1):
        self.stream.acquire()
        self.stream.write("\n" * count)
        self.stream.release()

    # gets overwritten by "master" attribute...
    # def master(self, level, *args):
    #     self.log(level, args, master_only=True)

    def info(self, *args, **kwargs):
        self.log(self.INFO, args, **kwargs)

    def debug(self, *args, **kwargs):
        self.log(self.DEBUG, args, **kwargs)

    def warning(self, *args, **kwargs):
        self.log(self.WARNING, args, **kwargs)

    def error(self, *args, **kwargs):
        self.log(self.ERROR, args, **kwargs)

    def critical(self, *args, **kwargs):
        self.log(self.CRITICAL, args, **kwargs)

    def exception(self, *args, **kwargs):
        self.log(self.ERROR, args, stack=traceback.format_exc(), **kwargs)

    def __str__(self):
        output = self.filename
        if not output:
            output = "stdout"
        return "RestLogger<{} {}>".format(self.name, output)


class ColorFormatter(logging.Formatter):
    def __init__(self, fmt=None, datefmt=None):
        if not fmt:
            fmt = "%(asctime)s - %(levelname)s - {YELLOW}%(name)s:%(threadName)s:{OFF} %(message)s".format(
                **ConsoleColors.__dict__
            )
        super(ColorFormatter, self).__init__(fmt, datefmt)

    def format(self, record):
        if record.levelname == "ERROR":
            return "\n{}{}{}\n".format(
                ConsoleColors.YELLOW, super(ColorFormatter, self).format(record), ConsoleColors.OFF
            )
        if record.levelname in LEVEL_COLORS:
            color = LEVEL_COLORS[record.levelname]
            record.levelname = "{}{}{}".format(
                ConsoleColors.__dict__[color], record.levelname, ConsoleColors.OFF
            )
        return super(ColorFormatter, self).format(record)

    def formatException(self, exc_info):
        return "{}{}{}".format(
            ConsoleColors.RED,
            super(ColorFormatter, self).formatException(exc_info),
            ConsoleColors.OFF,
        )


class BridgeLogFormatter(logging.Formatter):
    def format(self, record):
        return {
            "name": record.name,
            "message": record.msg,
            "asctime": record.asctime,
            "level": record.levelname,
            "func": record.funcName,
            "lineno": record.lineno,
            "pathname": record.pathname,
        }


from decimal import Decimal

PRETTY_INDENT = 4
PRETTY_MAX_VALUE_LENGTH = 6000
PRETTY_MAX_LINES = 100000
PRETTY_MAX_LENGTH = 500000


class ConsoleColors:
    BLACK = "\033[30m"
    RED = "\033[31m"
    GREEN = "\033[32m"
    YELLOW = "\033[33m"
    PINK = "\033[35m"
    BLUE = "\033[34m"
    WHITE = "\033[37m"

    HBLACK = "\033[90m"
    HRED = "\033[91m"
    HGREEN = "\033[92m"
    HYELLOW = "\033[93m"
    HBLUE = "\033[94m"
    HPINK = "\033[95m"
    HWHITE = "\033[97m"

    HEADER = "\033[95m"
    FAIL = "\033[91m"
    OFF = "\033[0m"
    BOLD = "\033[1m"
    UNDERLINE = "\033[4m"


def pp(*args):
    for arg in args:
        prettyWrite(arg)


def prettyWriteToString(d):
    output = StringIO()
    prettyWrite(d, output, colors=True)
    out = output.getvalue()
    output.close()
    return out


def dictToString(d):
    output = StringIO()
    prettyWrite(d, output, colors=True)
    out = output.getvalue()
    output.close()
    return out


def prettyLog(msg, data, logger=None):
    if not logger:
        logger = getLogger("root")
    out = prettyWriteToString(data)
    logger.info("{}\n{}".format(msg, out))


def prettyWrite(
    d, f=None, indent=PRETTY_INDENT, banner=None, line_count=0, ignore_false=False, colors=True
):
    std_output = False
    if f is None:
        std_output = True
        f = StringIO()

    if isinstance(d, dict):
        try:
            d = OrderedDict(sorted(d.items()))
        except:
            print("unabled to sort dict")
    prev = None
    if banner:
        f.write("---- BEGIN {} ----\n".format(banner))
    if type(d) is list:
        prev = False
        f.write("[")
        for i in d:
            if prev:
                line_count += 1
                f.write(",")
            else:
                line_count += 1
                # f.write('\n')
            flen = 0
            if hasattr(f, "len"):
                flen = f.len
            else:
                flen = f.tell()
            if line_count > PRETTY_MAX_LINES or flen >= PRETTY_MAX_LENGTH:
                f.write('{}"...truncated:PRETTY_MAX_LINES"'.format(" " * indent))
                break
            prev = True
            if type(i) is bool:
                i = int(i)
            if type(i) in [str, str]:
                if len(i) >= PRETTY_MAX_VALUE_LENGTH:
                    f.write(
                        '{}"{}...truncated:PRETTY_MAX_VALUE_LENGTH..."'.format(
                            " " * 1, i[: PRETTY_MAX_VALUE_LENGTH - 20]
                        )
                    )
                else:
                    f.write('{}"{}"'.format(" " * 1, i))
            elif type(i) is list or isinstance(i, dict):
                f.write(" " * (1))
                line_count = prettyWrite(i, f, indent + PRETTY_INDENT, line_count=line_count)
            elif type(i) is Decimal:
                f.write("{}{}".format(" " * 1, str(i)))
            else:
                f.write("{}{}".format(" " * 1, i))
        line_count += 1
        f.write("]")
    elif isinstance(d, dict):
        f.write("{")
        for key, value in list(d.items()):
            if ignore_false and not value:
                continue
            if prev:
                line_count += 1
                f.write(",\n")
            else:
                line_count += 1
                f.write("\n")

            if hasattr(f, "len"):
                flen = f.len
            else:
                flen = f.tell()
            if line_count > PRETTY_MAX_LINES or flen >= PRETTY_MAX_LENGTH:
                f.write('{}"truncated":"...truncated:PRETTY_MAX_LINES"\n'.format(" " * indent))
                break
            prev = True
            if type(key) in [str, str]:
                if colors:
                    f.write(
                        '{}{}"{}"{}:'.format(
                            " " * indent, ConsoleColors.YELLOW, key, ConsoleColors.OFF
                        )
                    )
                else:
                    f.write('{}"{}":'.format(" " * indent, key))
            else:
                f.write("{}{}: ".format(" " * indent, str(key)))
            if type(value) is list or isinstance(value, dict):
                f.write(" ")
                line_count = prettyWrite(value, f, indent + PRETTY_INDENT, line_count=line_count)
            else:
                if type(value) == str:
                    try:
                        if len(value) >= PRETTY_MAX_VALUE_LENGTH:
                            f.write(
                                ' "{}...truncated:PRETTY_MAX_VALUE_LENGTH"'.format(
                                    value[: PRETTY_MAX_VALUE_LENGTH - 20]
                                )
                            )
                        else:
                            if colors:
                                f.write(
                                    ' {}"{}"{}'.format(ConsoleColors.GREEN, value, ConsoleColors.OFF)
                                )
                            else:
                                f.write(' "{}"'.format(value))
                    except:
                        f.write(' "{}"'.format(repr(value)))
                elif type(value) in [datetime, time]:
                    if colors:
                        f.write(' {}"{}"{}'.format(ConsoleColors.GREEN, value, ConsoleColors.OFF))
                    else:
                        f.write(' "{}"'.format(value))
                elif type(value) is float:
                    if colors:
                        f.write(" {}{:f}{}".format(ConsoleColors.RED, value, ConsoleColors.OFF))
                    else:
                        f.write(" {:f}".format(value))
                elif type(value) is bool:
                    if colors:
                        if value:
                            f.write(" {}{}{}".format(ConsoleColors.GREEN, value, ConsoleColors.OFF))
                        else:
                            f.write(" {}{}{}".format(ConsoleColors.RED, value, ConsoleColors.OFF))
                    else:
                        f.write(" {}".format(value))
                elif type(value) is Decimal:
                    if colors:
                        f.write(" {}{}{}".format(ConsoleColors.BLUE, value, ConsoleColors.OFF))
                    else:
                        f.write(" {}".format(value))
                elif isinstance(value, bytearray):
                    f.write(" {}".format(str(bytes(hexlify(value)), encoding="utf-8")))
                else:
                    if colors:
                        f.write(" {}{}{}".format(ConsoleColors.RED, value, ConsoleColors.OFF))
                    else:
                        f.write(" {}".format(value))
        line_count += 1
        f.write("\n")
        f.write(" " * (indent - PRETTY_INDENT))
        f.write("}")
    else:
        f.write(str(d))

    if banner:
        f.write("\n---- END {} ----\n".format(banner))
    if indent == PRETTY_INDENT:
        f.write("\n")
    if std_output:
        sys.stdout.write(f.getvalue())
        f.close()
    return line_count


def mkdir(path):
    try:
        os.makedirs(path)
    except OSError as exc:
        if exc.errno == errno.EEXIST and os.path.isdir(path):
            pass
        else:
            raise


def pruneFile(path, bytes_remaining=ROTATE_LEFT_OVER_BYTES):
    with open(path, "r") as stream:
        try:
            stream.seek(0, 2)
            size = stream.tell()
            # now lets go back X bytes and find first new line
            stream.seek(size - bytes_remaining)
            stream.readline()
            to_end = stream.read()
        except Exception:
            pass
    with open(path, "w") as stream:
        stream.write(to_end)
        stream.flush()




#!/usr/bin/python2.4
#
# Copyright (C) 2010 Google Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Errors for the library.

All exceptions defined by the library
should be defined in this file.
"""

__author__ = 'jcgregorio@google.com (Joe Gregorio)'


from medialib.youtube.oauth2client import util
from medialib.youtube.oauth2client.anyjson import simplejson


class Error(Exception):
  """Base error for this module."""
  pass


class HttpError(Error):
  """HTTP data was invalid or unexpected."""

  @util.positional(3)
  def __init__(self, resp, content, uri=None):
    self.resp = resp
    self.content = content
    self.uri = uri

  def _get_reason(self):
    """Calculate the reason for the error from the response content."""
    reason = self.resp.reason
    try:
      data = simplejson.loads(self.content)
      reason = data['error']['message']
    except (ValueError, KeyError):
      pass
    if reason is None:
      reason = ''
    return reason

  def __repr__(self):
    if self.uri:
      return '<HttpError %s when requesting %s returned "%s">' % (
          self.resp.status, self.uri, self._get_reason().strip())
    else:
      return '<HttpError %s "%s">' % (self.resp.status, self._get_reason())

  __str__ = __repr__


class InvalidJsonError(Error):
  """The JSON returned could not be parsed."""
  pass


class UnknownFileType(Error):
  """File type unknown or unexpected."""
  pass


class UnknownLinkType(Error):
  """Link type unknown or unexpected."""
  pass


class UnknownApiNameOrVersion(Error):
  """No API with that name and version exists."""
  pass


class UnacceptableMimeTypeError(Error):
  """That is an unacceptable mimetype for this operation."""
  pass


class MediaUploadSizeError(Error):
  """Media is larger than the method can accept."""
  pass


class ResumableUploadError(HttpError):
  """Error occured during resumable upload."""
  pass


class InvalidChunkSizeError(Error):
  """The given chunksize is not valid."""
  pass

class InvalidNotificationError(Error):
  """The channel Notification is invalid."""
  pass

class BatchError(HttpError):
  """Error occured during batch operations."""

  @util.positional(2)
  def __init__(self, reason, resp=None, content=None):
    self.resp = resp
    self.content = content
    self.reason = reason

  def __repr__(self):
      return '<BatchError %s "%s">' % (self.resp.status, self.reason)

  __str__ = __repr__


class UnexpectedMethodError(Error):
  """Exception raised by RequestMockBuilder on unexpected calls."""

  @util.positional(1)
  def __init__(self, methodId=None):
    """Constructor for an UnexpectedMethodError."""
    super(UnexpectedMethodError, self).__init__(
        'Received unexpected call %s' % methodId)


class UnexpectedBodyError(Error):
  """Exception raised by RequestMockBuilder on unexpected bodies."""

  def __init__(self, expected, provided):
    """Constructor for an UnexpectedMethodError."""
    super(UnexpectedBodyError, self).__init__(
        'Expected: [%s] - Provided: [%s]' % (expected, provided))



# Generated by Django 4.1.4 on 2023-08-04 04:52

from django.db import migrations, models
import django.db.models.deletion
import rest.models.base
import rest.models.metadata


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('account', '0013_memberdevice_ip'),
        ('medialib', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='Page',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('modified', models.DateTimeField(auto_now=True, db_index=True)),
                ('order', models.IntegerField(blank=True, default=0)),
                ('title', models.CharField(max_length=255)),
                ('path', models.CharField(db_index=True, max_length=255)),
                ('slug', models.SlugField()),
                ('body', models.TextField(blank=True)),
                ('is_active', models.BooleanField(blank=True, default=True)),
                ('group', models.ForeignKey(blank=True, default=None, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='+', to='account.group')),
                ('member', models.ForeignKey(blank=True, default=None, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='+', to='account.member')),
                ('parent', models.ForeignKey(blank=True, default=None, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='children', to='wiki.page')),
            ],
            bases=(models.Model, rest.models.base.RestModel, rest.models.metadata.MetaDataModel),
        ),
        migrations.CreateModel(
            name='PageMetaData',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('category', models.CharField(blank=True, db_index=True, default=None, max_length=32, null=True)),
                ('key', models.CharField(db_index=True, max_length=80)),
                ('value_format', models.CharField(max_length=16)),
                ('value', models.TextField()),
                ('int_value', models.IntegerField(blank=True, default=None, null=True)),
                ('float_value', models.IntegerField(blank=True, default=None, null=True)),
                ('parent', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='properties', to='wiki.page')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='PageMedia',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('entry', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='media', to='wiki.page')),
                ('group', models.ForeignKey(blank=True, default=None, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='+', to='account.group')),
                ('media', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='+', to='medialib.mediaitem')),
                ('member', models.ForeignKey(blank=True, default=None, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='+', to='account.member')),
            ],
            bases=(models.Model, rest.models.base.RestModel),
        ),
    ]



# Generated by Django 4.1.4 on 2024-02-23 20:55

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('location', '0004_remove_address_modified_by_address_group_and_more'),
        ('account', '0018_userpasskey'),
    ]

    operations = [
        migrations.AddField(
            model_name='group',
            name='location',
            field=models.ForeignKey(blank=True, default=None, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='+', to='location.address'),
        ),
    ]



from ..engines import video_getinfo
from ..render_utils import *
from ..exceptions import *

import os
import subprocess
import tempfile

PresetConfig = {
    'name':		'Validate Video',
    'description':	'Basic video validation',
    'stage':	1,
    'default_use':	'validate',
    'applicable_kinds': 'V',
    'parameters': {
        'min_duration': {
            'description':	'Minimum video duration',
            'required':	False,
            'kind':		'F',
            'configurable':	True,
        },
        'max_duration': {
            'description':	'Maximum video duration',
            'required':	False,
            'kind':		'F',
            'configurable':	True,
        },
    },
}

def render(item, renditiondef, params):
    original = get_rendition(item, 'original')

    min_duration = params.get('min_duration', 0)
    max_duration = params.get('max_duration', 0)

    if min_duration < 60:
        min_duration_str = "%d seconds" % (min_duration)
    elif min_duration < 3600:
        if min_duration % 60 >= 55:
            min_duration_str = "%d minutes" % (min_duration/60 + 1)
        elif min_duration % 60 < 1:
            min_duration_str = "%d minutes" % (min_duration/60)
        else:
            min_duration_str = "%d minutes %d seconds" % (min_duration/60, min_duration%60)
    else:
        if min_duration % 3600 >= 3300:
            min_duration_str = "%d hours" % (min_duration/60 + 1)
        elif min_duration % 3600 < 60:
            min_duration_str = "%d hours" % (min_duration/60)
        else:
            min_duration_str = "%d hours %d minutes" % (min_duration/3600, min_duration%3600)

    if max_duration < 60:
        max_duration_str = "%d seconds" % (max_duration)
    elif max_duration < 3600:
        if max_duration % 60 < 5:
            max_duration_str = "%d minutes" % (max_duration/60)
        else:
            max_duration_str = "%d minutes %d seconds" % (max_duration/60, max_duration%60)
    else:
        if max_duration % 3600 < 300:
            max_duration_str = "%d hours" % (max_duration/60)
        else:
            max_duration_str = "%d hours %d minutes" % (max_duration/3600, max_duration%3600)

    meta = video_getinfo.getinfo(original.filename, params)
    if not (meta and ('duration' in meta or 'height' in meta)):
        raise ValidateError("Cannot decode uploaded file")
    if 'duration' in meta and min_duration and meta.get('duration', 0) < min_duration:
        raise ValidateError("Videos must be longer than {0} seconds and is {1}".format(min_duration, meta.get('duration', 0)))
    if 'duration' in meta and max_duration and meta.get('duration', 0) > max_duration:
        raise ValidateError("Videos must be shorter than {0} minutes".format(max_duration/60))
    if meta.get('video_codec', "").startswith("Apple ProRes 422"):
        raise ValidateError("Apple ProRes codec is currently not supported.  Please re-encode your video.")
    if meta.get('audio_channels', 2) > 2:
        raise ValidateError("Maximum of 2 audio channels allowed.")

    if 'height' in meta:
        original.height = meta.pop('height')
    if 'width' in meta:
        original.width = meta.pop('width')
    
    save_meta(original, meta)
    original.save()



# Copyright (C) 2012 Google Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utilities for Google Compute Engine

Utilities for making it easier to use OAuth 2.0 on Google Compute Engine.
"""

__author__ = 'jcgregorio@google.com (Joe Gregorio)'

import httplib2
import logging
import uritemplate

from oauth2client import util
from oauth2client.anyjson import simplejson
from oauth2client.client import AccessTokenRefreshError
from oauth2client.client import AssertionCredentials

logger = logging.getLogger(__name__)

# URI Template for the endpoint that returns access_tokens.
META = ('http://metadata.google.internal/0.1/meta-data/service-accounts/'
        'default/acquire{?scope}')


class AppAssertionCredentials(AssertionCredentials):
  """Credentials object for Compute Engine Assertion Grants

  This object will allow a Compute Engine instance to identify itself to
  Google and other OAuth 2.0 servers that can verify assertions. It can be used
  for the purpose of accessing data stored under an account assigned to the
  Compute Engine instance itself.

  This credential does not require a flow to instantiate because it represents
  a two legged flow, and therefore has all of the required information to
  generate and refresh its own access tokens.
  """

  @util.positional(2)
  def __init__(self, scope, **kwargs):
    """Constructor for AppAssertionCredentials

    Args:
      scope: string or iterable of strings, scope(s) of the credentials being
        requested.
    """
    self.scope = util.scopes_to_string(scope)

    # Assertion type is no longer used, but still in the parent class signature.
    super(AppAssertionCredentials, self).__init__(None)

  @classmethod
  def from_json(cls, json):
    data = simplejson.loads(json)
    return AppAssertionCredentials(data['scope'])

  def _refresh(self, http_request):
    """Refreshes the access_token.

    Skip all the storage hoops and just refresh using the API.

    Args:
      http_request: callable, a callable that matches the method signature of
        httplib2.Http.request, used to make the refresh request.

    Raises:
      AccessTokenRefreshError: When the refresh fails.
    """
    uri = uritemplate.expand(META, {'scope': self.scope})
    response, content = http_request(uri)
    if response.status == 200:
      try:
        d = simplejson.loads(content)
      except Exception as e:
        raise AccessTokenRefreshError(str(e))
      self.access_token = d['accessToken']
    else:
      raise AccessTokenRefreshError(content)



# Generated by Django 4.1.1 on 2022-09-24 16:22

from django.db import migrations
import rest.fields


class Migration(migrations.Migration):

    dependencies = [
        ('account', '0008_memberdevice_memberdevicemetadata'),
    ]

    operations = [
        migrations.AlterField(
            model_name='member',
            name='phone_number',
            field=rest.fields.FormattedField(db_index=True, default=None, max_length=64, null=True),
        ),
    ]



from rest import decorators as rd
from rest.views import restPermissionDenied, restStatus, restHTML
from rest import helpers as rh
from account import models as am


@rd.url(r'^member$')
@rd.url(r'^member/$')
@rd.url(r'^member/(?P<pk>\d+)$')
@rd.login_required
def rest_on_member(request, pk=None):
    return am.Member.on_rest_request(request, pk)


@rd.urlPOST('member/motd/reset')
@rd.perm_required('manage_members')
def rest_on_clear_motd(request):
    key = request.DATA.get("key")
    if key.startswith("viewed."):
        key = key.split('.')[-1]
        am.MemberMetaData.objects.filter(key=key, category="viewed").delete()
    else:
        am.MemberMetaData.objects.filter(key="saw_motd", category__isnull=True).delete()
    return restStatus(request, True)


@rd.url(r'^member/me$')
@rd.login_optional
def member_me_action(request):
    if not request.user.is_authenticated:
        return restPermissionDenied(request, "not authenticated")
    if request.method == "GET":
        # request.session['ws4redis:memberof'] = request.member.getGroupUUIDs()
        # from rest import helpers as rh
        # rh.debug("user_platform", request.DATA.getUserAgentPlatform())
        return request.member.on_rest_get(request)
    elif request.method == "POST":
        return request.member.on_rest_post(request)
    return restStatus(request, False, error="not supported")


@rd.url(r'^authtoken$')
@rd.url(r'^authtoken/(?P<pk>\d+)$')
@rd.login_required
def rest_on_authtoken(request, pk=None):
    return am.AuthToken.on_rest_request(request, pk)


@rd.url(r'^session$')
@rd.url(r'^session/(?P<pk>\d+)$')
@rd.login_required
def rest_on_session(request, pk=None):
    return am.AuthSession.on_rest_request(request, pk)


@rd.urlGET('unsubscribe')
@rd.requires_params(["t"])
def rest_on_member(request):
    t = request.DATA.get("t")
    m = am.Member.objects.filter(uuid=t).last()
    if m is not None:
        m.addPermission("email_disabled")
        m.reportIncident("email", f"{m.email} has unsubscribed to all email")
    context = rh.getContext(request, member=m)
    if request.DATA.get("format") == "json":
        return restStatus(request, True)
    return restHTML(request, template="unsubscribed.html", context=context)



# Generated by Django 4.1.3 on 2022-11-28 17:18

from django.db import migrations, models
import django.db.models.deletion
import rest.models


class Migration(migrations.Migration):

    dependencies = [
        ('account', '0011_authtoken'),
    ]

    operations = [
        migrations.CreateModel(
            name='Settings',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('modified', models.DateTimeField(auto_now=True)),
                ('name', models.CharField(max_length=128, unique=True)),
                ('group', models.ForeignKey(default=None, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='settings', to='account.group')),
                ('owner', models.ForeignKey(default=None, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='settings', to='account.member')),
            ],
            bases=(models.Model, rest.models.RestModel, rest.models.MetaDataModel),
        ),
        migrations.CreateModel(
            name='SettingsMetaData',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('category', models.CharField(blank=True, db_index=True, default=None, max_length=32, null=True)),
                ('key', models.CharField(db_index=True, max_length=80)),
                ('value_format', models.CharField(max_length=16)),
                ('value', models.TextField()),
                ('int_value', models.IntegerField(blank=True, default=None, null=True)),
                ('float_value', models.IntegerField(blank=True, default=None, null=True)),
                ('parent', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='properties', to='account.settings')),
            ],
            options={
                'abstract': False,
            },
        ),
    ]



from django.db import models
from rest import models as rm


class GeoTrack(models.Model, rm.RestModel):
    uid = models.CharField(max_length=124, db_index=True, null=True, default=None)
    owner = models.ForeignKey("account.User", null=True, default=None, on_delete=models.CASCADE)
    name = models.CharField(max_length=124)
    kind = models.CharField(max_length=64)
    ref_id = models.IntegerField(blank=True, default=0)
    created = models.DateTimeField(auto_now_add=True)

    def getPositions(self):
        return self.positions.all().order_by("created")

    def getCurrentPosition(self):
        positions = self.getPositions()
        if positions.count():
            return positions[0]
        return None


class GeoPosition(models.Model, rm.RestModel):
    uid = models.CharField(max_length=124, db_index=True, null=True, default=None)
    track = models.ForeignKey(GeoTrack, blank=True, null=True, default=None, related_name="positions", on_delete=models.CASCADE)
    created = models.DateTimeField(auto_now_add=True)
    lat = models.FloatField(default=0.0)
    lng = models.FloatField(default=0.0)
    elevation = models.FloatField(default=0.0, blank=True)
    accuracy = models.FloatField(default=0.0, blank=True)
    altitude = models.FloatField(default=0.0, blank=True)
    altitudeAccuracy = models.FloatField(default=0.0, blank=True)
    heading = models.FloatField(default=0.0, blank=True)
    speed = models.FloatField(default=0.0, blank=True)



# Generated by Django 4.1.4 on 2023-01-06 22:15

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('incident', '0005_incident_component_alter_incident_state'),
    ]

    operations = [
        migrations.DeleteModel(
            name='EventMetaData',
        ),
    ]



from ..exceptions import *
from ..render_utils import RenderProc
from . import video_getinfo

from django import db

from medialib import stores
from medialib.models import RenditionSegment

import os
import tempfile
import subprocess
import shutil
import glob
import signal
import sys
import errno

hls_cmd = ['streamsegmenter', "-s", "120", "-r", "3600"]

def sigusr(signum, frame):
    pass

def sigterm(proc):
    def sighand(signum, frame):
        try:
            proc.terminate()
        except OSError:
            pass

    return sighand

def _collect(tmpdir, url, rendition):
    while True:
        segments = glob.glob(os.path.join(tmpdir, "segment-*.ts"))
        if not segments:
            break
        segints = []
        for seg in segments:
            fn = os.path.split(seg)[-1]
            segnum = int(fn[8:-3])
            segints.append((segnum, seg))
        segints.sort()
        for (segnum, seg) in segints:
            fp = open(seg)
            segurl = url + str(segnum) + ".ts"
            stores.upload(segurl, fp)
            meta = video_getinfo.getinfo(seg)
            fsize = os.fstat(fp.fileno()).st_size
            rendition.bytes += fsize
            rendition.__class__.objects.filter(pk=rendition.pk).update(bytes=rendition.bytes)
            rseg = RenditionSegment(rendition=rendition, segment=segnum, start=meta['start'], duration=meta['duration'], end=meta['start']+meta['duration'], bytes=fsize, url=segurl)
            rseg.save()
            fp.close()
            rendition.set_meta("last_segment", segnum)
            os.unlink(seg)


def run(rendition, infile, params={}, extra_args=[], final_args=[]):
    cmd = list(hls_cmd)
    
    copy_params = (	('segment_duration', '-d'),
              )
    for p in copy_params:
        if p[0] in params:
            s = str(params[p[0]])
            if len(p) > 2:
                s = p[2] + s
            if len(p) > 3:
                s = s + p[3]
            cmd += [p[1], s]

    tmpdir = tempfile.mkdtemp()

    (rp, wp) = os.pipe()
    pid = os.fork()
    if pid == 0:
        os.dup2(wp, sys.stdout.fileno())
        os.dup2(wp, sys.stderr.fileno())
        os.close(rp)
        os.close(wp)
        db.connection.close()
        pid = os.getpid()
        cmd += ["-p", str(pid)] + extra_args + final_args + [tmpdir, "segment", '']

        signal.signal(signal.SIGUSR1, sigusr)

        fd = os.open(infile, os.O_RDONLY)
        proc = subprocess.Popen(cmd, close_fds=True, stdin=fd)
        os.close(fd)
        signal.signal(signal.SIGTERM, sigterm(proc))

        rval = 250
        try:
            while True:
                try:
                    pid, sts = os.waitpid(proc.pid, 0)
                    if pid >= 0:
                        rval = os.WEXITSTATUS(sts)
                        break
                except OSError as err:
                    if err.errno == errno.EINTR:
                        _collect(tmpdir, rendition.url, rendition)
                        db.reset_queries()
                        continue
                    raise
                except:
                    raise
        finally:
            try:
                _collect(tmpdir, rendition.url, rendition)
            except:
                pass
            rendition.set_meta("final", "1")
            shutil.rmtree(tmpdir)
        os._exit(rval)
    else:
        os.close(wp)
        return RenderProc(pid, errfile=rp)




from datetime import datetime
import time
from rest import helpers
from django.conf import settings
import django
import version
import platform
import socket
import subprocess


try:
    import psutil
except Exception:
    print("no psutil")

SOFTWARE_VERSIONS = getattr(settings, 'SOFTWARE_VERSIONS', None)


def getHostIP():
    hostname = socket.gethostname()
    return socket.gethostbyname(hostname)


def getTcpEstablishedCount():
    return len(getTcpEstablished())


def getTcpEstablished(filter=None):
    cons = psutil.net_connections(kind="tcp")
    established = []
    for c in cons:
        if c.status == "ESTABLISHED":
            established.append(c)
    return filterConnections(established, filter)


def filterConnections(cons, filter):
    if filter == "https":
        return [c for c in cons if c.laddr.port == 443]
    elif filter == "redis":
        return [c for c in cons if c.raddr.port == 6379]
    elif filter == "postgres":
        return [c for c in cons if c.raddr.port == 5432]
    elif filter == "unknown":
        return [c for c in cons if c.raddr.port not in [5432, 6379] and c.laddr.port != 443]
    elif filter and ":" in filter:
        addr, port = filter.split(':')
        if addr == "raddr":
            return [c for c in cons if c.raddr.port == port]
        elif addr == "laddr":
            return [c for c in cons if c.laddr.port == port]
    return cons


def consToDict(cons):
    out = []
    cid = 0
    for c in cons:
        cid += 1
        out.append({
            "id": cid,
            "type": c.type.name,
            "status": c.status,
            "family": c.family.name,
            "raddr": {
                "port": c.raddr.port,
                "ip": c.raddr.ip
            },
            "laddr": {
                "port": c.laddr.port,
                "ip": c.laddr.ip
            }
        })
    return out


def getVersions(out=None):
    if out is None:
        out = {}
    for key in SOFTWARE_VERSIONS:
        if key == "django":
            out[key] = django.__version__
        else:
            out[key] = safe_cmd(SOFTWARE_VERSIONS[key])
    return out


def getHostInfo(include_versions=False, include_blocked=False):
    mem = psutil.virtual_memory()
    disk = psutil.disk_usage('/')
    net = psutil.net_io_counters()

    out = {
        "time": time.time(),
        "datetime": str(datetime.now()),
        "version": version.VERSION,
        "os": {
            "system": platform.system(),
            "version": platform.version(),
            "hostname": platform.node(),
            "release": platform.release(),
            "processor": platform.processor(),
            "machine": platform.machine()
        },
        "boot_time": psutil.boot_time(),
        "cpu_load": psutil.cpu_percent(),
        "cpus_load": psutil.cpu_percent(percpu=True),
        "memory": {
            "total": mem.total,
            "used": mem.used,
            "available": mem.available,
            "percent": mem.percent
        },
        "disk": {
            "total": disk.total,
            "used": disk.used,
            "free": disk.free,
            "percent": disk.percent
        }, 
        "users": psutil.users()
    }

    try:
        out["cpu"] = {
            "count": psutil.cpu_count(),
            "freq": psutil.cpu_freq(),
        }
    except Exception:
        pass

    try:
        out["network"] = {
            "tcp_cons": getTcpEstablishedCount(),
            "bytes_sent": net.bytes_sent,
            "bytes_recv": net.bytes_recv,
            "packets_sent": net.packets_sent,
            "packets_recv": net.packets_recv,
            "errin": net.errin,
            "errout": net.errout,
            "dropin": net.dropin,
            "dropout": net.dropout
        }
    except Exception:
        pass

    if include_versions and SOFTWARE_VERSIONS:
        out["versions"] = getVersions()
    if include_blocked:
        out["blocked"] = helpers.getBlockedHosts()
    return out


def safe_cmd(cmd, *args):
    try:
        cmd_args = [cmd]
        if len(args):
            cmd_args.extend(list(args))
        return helpers.toString(subprocess.check_output(cmd_args, shell=True).strip())
    except Exception as err:
        return str(err)
        # print( str(err))
    return None



# Generated by Django 4.1.4 on 2024-04-15 13:45

from django.db import migrations, models
import django.db.models.deletion
import rest.fields
import rest.models.base
import rest.models.metadata


class Migration(migrations.Migration):

    dependencies = [
        ('account', '0019_group_location'),
    ]

    operations = [
        migrations.CreateModel(
            name='CloudCredentials',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('modified', models.DateTimeField(auto_now=True)),
                ('name', models.CharField(blank=True, default=None, max_length=128, null=True)),
                ('uuid', models.CharField(blank=True, db_index=True, default=None, max_length=64, null=True)),
                ('state', models.IntegerField(db_index=True, default=1)),
                ('credentials', rest.fields.JSONField(default=None, null=True)),
                ('group', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='cloud_credentials', to='account.group')),
            ],
            bases=(models.Model, rest.models.base.RestModel, rest.models.metadata.MetaDataModel),
        ),
        migrations.CreateModel(
            name='CloudCredentialsMetaData',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('category', models.CharField(blank=True, db_index=True, default=None, max_length=32, null=True)),
                ('key', models.CharField(db_index=True, max_length=80)),
                ('value_format', models.CharField(max_length=16)),
                ('value', models.TextField()),
                ('int_value', models.IntegerField(blank=True, default=None, null=True)),
                ('float_value', models.IntegerField(blank=True, default=None, null=True)),
                ('parent', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='properties', to='account.cloudcredentials')),
            ],
            options={
                'abstract': False,
            },
        ),
    ]



from rest import settings
from objict import objict
# import boto
# from boto.s3.connection import S3Connection
import boto3
import botocore

from urllib.parse import urlparse
import io
import sys
from medialib import utils
import threading
import tempfile

S3 = objict(
    KEY=settings.AWS_KEY,
    SECRET=settings.AWS_SECRET,
    REGSION=settings.AWS_REGION,
    BUCKET=settings.AWS_S3_BUCKET)


class S3Item(object):
    def __init__(self, url, content_type="application/json", metadata=None):
        self.url = url
        u = urlparse(url)
        self.bucket_name = u.netloc
        self.key = u.path.lstrip('/')
        self.content_type = content_type
        self.s3 = getS3()
        self.host = "https://s3.amazonaws.com"
        self.metadata = metadata
        self.object = getObject(self.bucket_name, self.key)
        self.exists = self.checkExists()

    def checkExists(self):
        try:
            self.object.load()
        except botocore.exceptions.ClientError as e:
            if e.response['Error']['Code'] == "404":
                # The object does not exist.
                return False
            else:
                # Something else has gone wrong.
                raise
        return True

    def upload(self, fp, background=False):
        if self.metadata:
            return self.object.upload_fileobj(openFile(fp), ExtraArgs=self.metadata)
        self.object.upload_fileobj(openFile(fp))

    def createUpload(self):
        self.part_num = 0
        self.parts = []
        self.client = getS3(False)
        resp = self.client.create_multipart_upload(Bucket=self.bucket_name, Key=self.key)
        self.upload_id = resp["UploadId"]
        return True

    def uploadPart(self, chunk):
        self.part_num += 1
        resp = self.client.upload_part(
            Bucket=self.bucket_name,
            Key=self.key,
            PartNumber=self.part_num,
            UploadId=self.upload_id,
            Body=chunk)
        self.parts.append(dict(PartNumber=self.part_num, ETag=resp["ETag"]))

    def completeUpload(self):
        resp = self.client.complete_multipart_upload(
            Bucket=self.bucket_name,
            Key=self.key,
            UploadId=self.upload_id,
            MultipartUpload=dict(Parts=self.parts))
        return resp

    @property
    def public_url(self):
        return "{}/{}/{}".format(self.host, self.bucket_name, self.key)

    def generateURL(self, expires=600):
        client = getS3(False)
        return client.generate_presigned_url(
            'get_object',
            ExpiresIn=expires,
            Params={'Bucket': self.bucket_name, 'Key': self.key})

    def download(self, fp=None):
        if fp is None:
            fp = tempfile.NamedTemporaryFile()
        self.object.download_fileobj(fp)
        return fp

    def delete(self):
        self.object.delete()


def _getS3(as_resource=True, key=S3.KEY, secret=S3.SECRET):
    if as_resource:
        return boto3.resource('s3', aws_access_key_id=key, aws_secret_access_key=secret)
    return boto3.client('s3', aws_access_key_id=key, aws_secret_access_key=secret)


def getS3(as_resource=True):
    if as_resource:
        if S3.resource is None:
            S3.resource = _getS3(True)
        return S3.resource
    if S3.client is None:
        S3.client = _getS3(False)
    return S3.client


def getBucket(name):
    s3r = getS3()
    return s3r.Bucket(name)


def getObject(bucket_name, key):
    s3r = getS3()
    return s3r.Object(bucket_name, key)


def getObjectContent(bucket_name, key):
    s3r = getS3(False)
    obj = s3r.get_object(Bucket=bucket_name, Key=key)
    return obj['Body'].read().decode('utf-8')


class ProgressPercentage(object):
    def __init__(self, size):
        self._size = size
        self._seen_so_far = 0
        self._lock = threading.Lock()

    def __call__(self, bytes_amount):
        # To simplify we'll assume this is hooked up
        # to a single filename.
        with self._lock:
            self._seen_so_far += bytes_amount
            percentage = (self._seen_so_far / self._size) * 100
            sys.stdout.write(
                "\r%s  %s / %s  (%.2f%%)" % (
                    self._filename, self._seen_so_far, self._size,
                    percentage))
            sys.stdout.flush()


def upload(url, fp, background=False):
    obj = S3Item(url)
    obj.upload(fp)


def view_url_noexpire(url, is_secure=False):
    obj = S3Item(url)
    return obj.public_url


def view_url(url, expires=600, is_secure=True):
    if expires is None:
        return view_url_noexpire(url, is_secure)
    obj = S3Item(url)
    return obj.generateURL(expires)


def exists(url):
    obj = S3Item(url)
    return obj.exists


def get_file(url, fp=None):
    obj = S3Item(url)
    return obj.download(fp)


def generate_upload_url(url, filetype, expires=3600, acl="public-read"):
    u = urlparse(url)
    bucket_name = u.netloc
    key = u.path.lstrip('/')
    client = getS3(False)
    params = dict(Bucket=bucket_name, Key=key, ContentType=filetype)
    from rest import helpers
    helpers.log_error("generate_upload_url", params)
    return client.generate_presigned_url(
        'put_object',
        ExpiresIn=expires,
        Params=params)


def delete(url):
    if url[-1] == "/":
        prefix = url.path.lstrip("/")
        bucket_name = url.netloc
        s3 = getS3()
        response = s3.list_objects_v2(
            Bucket=bucket_name,
            Prefix =prefix,
            MaxKeys=100)
        for obj in response['Contents']:
            s3.delete_object(Bucket=bucket_name, Key=object['Key'])
    else:
        obj = S3Item(url)
        obj.delete()
        # _getkey(url, key=settings.AWS_ADMIN_KEY, secret=settings.AWS_ADMIN_SECRET).delete()


def openFile(fp):
    # this should fail if already opened.. if not iwll open
    # even if wrapped file-like object exists. To avoid Django-specific
    # logic, pass a copy of internal file-like object if `content` is
    # `File` class instance.
    if isinstance(fp, (str, bytes)):
        return io.BytesIO(utils.toBytes(fp))
    if hasattr(fp, "read"):
        return io.BytesIO(utils.toBytes(fp.read()))
    try:
        return open(fp.name, "r")
    except IOError:
        pass
    return fp


def path(url):
    u = urlparse(url)
    return u.path



from importlib import import_module
from django.utils.cache import patch_vary_headers
from django.utils.http import http_date

from rest import ua
from rest import helpers
from rest import settings
from rest import crypto
import time

# HACK FOR SameSite
from http.cookies import Morsel
Morsel._reserved['samesite'] = 'SameSite'
SESSION_COOKIE_SAMESITE = settings.SESSION_COOKIE_SAMESITE


def sessionKeyFromRequest(request):
    token = request.META.get('HTTP_AUTHORIZATION', "").strip()
    if token.count(" ") == 1:
        bearer, session_key = token.split(" ")
        if bearer.lower == "bearer":
            session_key = session_key.split(".")[2]
        if len(session_key) > 111:
            session_key = session_key[-111:]
        # lets key it ip specific
        session_key = f"{session_key}{request.ip}"
    else:
        # no longer support cookies
        # security risk should be low as these are not authenticated sessions
        user_agent = request.META.get('HTTP_USER_AGENT', None)
        if user_agent is None:
            return request.COOKIES.get(settings.SESSION_COOKIE_NAME)
        session_key = crypto.hashSHA256(user_agent, request.ip)
    return session_key


class SimpleSessionMiddleware(object):
    def __init__(self, get_response):
        self.get_response = get_response
        engine = import_module(settings.SESSION_ENGINE)
        self.SessionStore = engine.SessionStore

    def __call__(self, request):
        try:
            self.process_request(request)
            response = self.get_response(request)
            self.process_response(request, response)
        except Exception:
            helpers.log_exception()
        return response

    def process_request(self, request):
        request.ip = helpers.getRemoteIP(request)
        if request.method != "OPTIONS":
            # if we are using auth header tokens, let us not set a session id
            session_key = sessionKeyFromRequest(request)
            request.session = self.SessionStore(session_key)
            request.location = None
        else:
            request.session = self.SessionStore()
        # helpers.log_error("session keys", session_key, request.session.session_key)

    def process_response(self, request, response):
        if response.status_code == 500:
            return response
        if request.session is None or request.method == "OPTIONS":
            return response
        # print("saving session: \n{}".format(request.session.session_key))
        # print("session cooking domain: {}".format(settings.SESSION_COOKIE_DOMAIN))
        # print("session cooking path: {}".format(settings.SESSION_COOKIE_PATH))
        # print("session max_age: {}".format(max_age))
        request.session.save()
        if not settings.ALLOW_COOKIES or hasattr(request, "token"):
            return response
        return self.save_session_to_cookie(request, response)
    
    def save_session_to_cookie(self, request, response):
        expires = None
        max_age = None
        session_key = request.COOKIES.get(settings.SESSION_COOKIE_NAME)
        if session_key == request.session.session_key:
            return response
        # helpers.log_error("new session keys", request.session.session_key)
        if request.session.accessed: 
            patch_vary_headers(response, ('Cookie',))
        response.set_cookie(
            settings.SESSION_COOKIE_NAME,
            request.session.session_key, max_age=max_age,
            expires=expires, domain=settings.SESSION_COOKIE_DOMAIN,
            path=settings.SESSION_COOKIE_PATH,
            secure=settings.SESSION_COOKIE_SECURE or None,
            httponly=settings.SESSION_COOKIE_HTTPONLY or None)
        if SESSION_COOKIE_SAMESITE:
            # now lets verify browser supports it
            if hasattr(request, "DATA"):
                ua_info = request.DATA.getUserAgentInfo()
            else:
                ua_info = ua.parse(request.META.get('HTTP_USER_AGENT', ''))
            family = ua_info.get("user_agent.family", None)
            if not family:
                return response

            major = ua_info.get("user_agent.major", 0)
            if bool(major):
                major = int(major)
                if family.lower() == "chrome" and major < 75:
                    return response
                elif family.lower() == "firefox" and major < 60:
                    return response
                elif family.lower() == "safari" and major < 12:
                    return response     
                elif family.lower() == "ie" and major < 16:
                    return response                       
            response.cookies[settings.SESSION_COOKIE_NAME]["samesite"] = SESSION_COOKIE_SAMESITE
        return response


class SessionMiddleware(object):
    def __init__(self, get_response):
        self.get_response = get_response
        engine = import_module(settings.SESSION_ENGINE)
        self.SessionStore = engine.SessionStore

    def __call__(self, request):
        self.process_request(request)
        response = self.get_response(request)
        self.process_response(request, response)
        return response

    def process_request(self, request):
        session_key = request.COOKIES.get(settings.SESSION_COOKIE_NAME)
        # for key in request.COOKIES:
        #     print("cookie.{} = {}".format(key, request.COOKIES.get(key)))
        # # print(session_key)
        # print("session key: {}".format(session_key))
        # print("session SESSION_COOKIE_NAME: {}".format(settings.SESSION_COOKIE_NAME))

        secure_keys = settings.get("SESSION_KEY_SECURE", True)
        if not secure_keys:
            if not session_key:
                session_key = request.META.get('HTTP_AUTHORIZATION', None)
                if session_key and session_key.count(" ") > 0:
                    # this is not a session key
                    session_key = None

            if not session_key:
                session_key = request.META.get('HTTP_X_SESSIONID', None)

            if not session_key:
                session_key = request.POST.get('SESSION_KEY', None)

            if not session_key:
                session_key = request.GET.get('SESSION_KEY', None)

        if session_key:
            request.session = self.SessionStore(session_key)
            if not request.session.exists(session_key):
                session_key = None
        request.session = self.SessionStore(session_key)
        # check if session key is dead?
        # print(request.session)

    def process_response(self, request, response):
        """
        If request.session was modified, or if the configuration is to save the
        session every time, save the changes and set a session cookie or delete
        the session cookie if the session has been emptied.
        """
        try:
            accessed = request.session.accessed
            modified = request.session.modified
            empty = request.session.is_empty()
        except AttributeError as err:
            # print("session attribute error: {}".format(str(err)))
            pass
        else:
            # First check if we need to delete this cookie.
            # The session should be deleted only if the session is entirely empty
            if settings.SESSION_COOKIE_NAME in request.COOKIES and empty:
                response.delete_cookie(
                    settings.SESSION_COOKIE_NAME,
                    domain=settings.SESSION_COOKIE_DOMAIN)
            else:
                if accessed:
                    patch_vary_headers(response, ('Cookie',))
                if (modified or settings.SESSION_SAVE_EVERY_REQUEST) and not empty:
                    if request.session.get_expire_at_browser_close():
                        max_age = None
                        expires = None
                    else:
                        max_age = request.session.get_expiry_age()
                        expires_time = time.time() + max_age
                        expires = http_date(expires_time)
                    # Save the session data and refresh the client cookie.
                    # Skip session save for 500 responses, refs #3881.
                    if response.status_code != 500:
                        # print("saving session: \n{}".format(request.session.session_key))
                        # print("session cooking domain: {}".format(settings.SESSION_COOKIE_DOMAIN))
                        # print("session cooking path: {}".format(settings.SESSION_COOKIE_PATH))
                        # print("session max_age: {}".format(max_age))
                        request.session.save()
                        response.set_cookie(
                            settings.SESSION_COOKIE_NAME,
                            request.session.session_key, max_age=max_age,
                            expires=expires, domain=settings.SESSION_COOKIE_DOMAIN,
                            path=settings.SESSION_COOKIE_PATH,
                            secure=settings.SESSION_COOKIE_SECURE or None,
                            httponly=settings.SESSION_COOKIE_HTTPONLY or None)
                        if SESSION_COOKIE_SAMESITE:
                            # now lets verify browser supports it
                            ua_info = ua.parse(request.META.get('HTTP_USER_AGENT', ''))
                            family = ua_info.get("user_agent.family", None)
                            if not family:
                                return response
                            major = int(ua_info.get("user_agent.major", 0))
                            if family.lower() == "chrome" and major < 75:
                                return response
                            elif family.lower() == "firefox" and major < 60:
                                return response
                            elif family.lower() == "safari" and major < 12:
                                return response     
                            elif family.lower() == "ie" and major < 16:
                                return response                       
                            response.cookies[settings.SESSION_COOKIE_NAME]["samesite"] = SESSION_COOKIE_SAMESITE
        return response





from rest import decorators as rd
from location import models as location


@rd.url('address')
@rd.url('address/<int:pk>')
def rest_on_address(request, pk=None):
    return location.Address.on_rest_request(request, pk)



from rest.decorators import periodic
from taskqueue.models import Task


@periodic(minute=50, hour=8)
def run_cleanup(force=False, verbose=False, now=None):
    Task.Publish("incident", "run_cleanup", channel="tq_app_handler")


@periodic(minute=30)
def run_auto_close(force=False, verbose=False, now=None):
    Task.Publish("incident", "run_auto_close", channel="tq_app_handler")



from rest import settings
import requests
from objict import objict

GEOCODE_URL = "https://maps.googleapis.com/maps/api/geocode/json"
PLACE_FROMTEXT_URL = "https://maps.googleapis.com/maps/api/place/findplacefromtext/json"
PLACE_DETAILS_URL = "https://maps.googleapis.com/maps/api/place/details/json"


def search(address, normalize=True):
    if settings.GOOGLE_MAPS_KEY is None:
        raise Exception("requires GOOGLE_MAPS_KEY")
    params = dict(address=address, key=settings.GOOGLE_MAPS_KEY)
    res = requests.get(GEOCODE_URL, params=params)
    if res.status_code != 200:
        return None
    if normalize:
        return normalizeResponse(res.json())
    return parseResponse(res.json())


def locationAt(lat, lng, normalize=True):
    if settings.GOOGLE_MAPS_KEY is None:
        raise Exception("requires GOOGLE_MAPS_KEY")
    params = dict(latlng=f"{lat},{lng}", key=settings.GOOGLE_MAPS_KEY)
    res = requests.get(GEOCODE_URL, params=params)
    if res.status_code != 200:
        return None
    if normalize:
        return normalizeResponse(res.json())
    return parseResponse(res.json())


def findPlaces(lat, lng, query="", radius_meters=50, normalize=True):
    if settings.GOOGLE_MAPS_KEY is None:
        raise Exception("requires GOOGLE_MAPS_KEY")
    params = dict(
        key=settings.GOOGLE_MAPS_KEY,
        input=query,
        inputtype="textquery",
        circular=f"circle:{radius_meters}@{lat,{lng}}")
    res = requests.get(GEOCODE_URL, params=params)
    if res.status_code != 200:
        return None
    if normalize:
        return normalizePlace(res.json())
    return parseResponse(res.json())


def parseResponse(response):
    locations = []
    results = response.get("results")
    for res in results:
        loc = objict.fromdict(res)
        locations.append(loc)
    return loc


GOOGLE_NORMALIZE_KEYS = {
    "locality": "city",
    "administrative_area_level_2": "county",
    "administrative_area_level_1": "state",
}


def normalizeResponse(response):
    locations = []
    results = response.get("results")
    for res in results:
        loc = objict(uuid=res.get("place_id", None), formatted=res.get("formatted_address", None))
        locations.append(loc)
        for comp in res.get("address_components"):
            key = comp.get("types", [None])[0]
            loc[GOOGLE_NORMALIZE_KEYS.get(key, key)] = comp.get("short_name")
        geometry = res.get("geometry", None)
        if geometry is not None:
            loc.update(geometry.get("location"))
    return locations


def normalizePlace(response):
    out = objict()
    out.name = response.name
    out.types = response.types
    out.place_id = response.place_id
    out.location = response.geometry.location
    out.photos = response.photos
    out.icon = response.icon
    out.address = response.formatted_address
    return out



from rest import decorators as rd
from .. import models as location


@rd.urlGET('ip')
def on_myip(request):
    gip = location.GeoIP.get(request.ip)
    graph = request.DATA.get("graph", "default")
    return gip.restGet(request, graph)


@rd.urlGET('ips')
def rest_on_geoip(request):
    return location.GeoIP.on_rest_list(request)


@rd.urlGET(r'^ip/lookup$')
def on_ip_lookup(request):
    ip = request.DATA.get("ip", request.ip)
    gip = location.GeoIP.get(
        ip,
        force_refresh=request.DATA.get("refresh", 0, field_type=bool),
        stale_after=request.DATA.get("stale_after", 90, field_type=int))
    graph = request.DATA.get("graph", "default")
    return gip.restGet(request, graph)


@rd.urlGET('geo/ip')
@rd.urlGET('geo/ip/<int:pk>')
def rest_on_geoip(request, pk=None):
    return location.GeoIP.on_rest_request(request, pk)


@rd.urlPOST('geo/ip/<int:pk>')
@rd.perm_required(["manage_location"])
def rest_save_geoip(request, pk=None):
    return location.GeoIP.on_rest_request(request, pk)


@rd.urlPOST('geo/ip')
@rd.login_required
@rd.requires_params(["ip"])
def rest_on_geoip_lookup(request, pk=None):
    ip = request.DATA.get("ip", request.ip)
    gip = location.GeoIP.get(
        ip,
        force_refresh=request.DATA.get("refresh", 0, field_type=bool),
        stale_after=request.DATA.get("stale_after", 90, field_type=int))
    graph = request.DATA.get("graph", "default")
    return gip.restGet(request, graph)



# Generated by Django 4.1.4 on 2023-11-07 12:44

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('account', '0013_memberdevice_ip'),
    ]

    operations = [
        migrations.AlterField(
            model_name='notificationmemberrecord',
            name='member',
            field=models.ForeignKey(default=None, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='notifications', to='account.member'),
        ),
    ]



from objict import objict
import requests


# THIS APPEARS TO BE DEAD?

def getLocation(ip):
    """
    https://extreme-ip-lookup.com/json/68.101.118.74
    {
     "businessName" : "",
     "businessWebsite" : "",
     "city" : "Dana Point",
     "continent" : "North America",
     "country" : "United States",
     "countryCode" : "US",
     "ipName" : "ip68-101-118-74.oc.oc.cox.net",
     "ipType" : "Residential",
     "isp" : "Cox Communications Inc.",
     "lat" : "33.46697",
     "lon" : "-117.69811",
     "org" : "Cox Communications Inc.",
     "query" : "68.101.118.74",
     "region" : "California",
     "status" : "success"
    }
    """
    # params = {"token":IPINFO_TOKEN}
    url = f"https://extreme-ip-lookup.com/json/{ip}"
    res = requests.get(url)
    if res.status_code != 200:
        return None
    data = objict.fromdict(res.json())

    loc = objict(provider="extreme api")
    loc.isp = data.isp
    loc.ip = data.ip
    loc.hostname = data.ipName
    loc.ip_type = data.ipType
    loc.latitude = data.lat
    loc.longitude = data.lon
    loc.continent = data.continent
    loc.region = data.region
    loc.country = data.country
    loc.countryCode = data.countryCode
    loc.state = data.region
    loc.city = data.city
    loc.business = data.businessName
    loc.website = data.businessWebsite
    return loc



# coding: utf-8

"""
    mistune_contrib.toc
    ~~~~~~~~~~~~~~~~~~~

    Support TOC features for mistune.

    :copyright: (c) 2015 by Hsiaoming Yang.
"""


class TocMixin(object):
    """TOC mixin for Renderer, mix this with Renderer::

        class TocRenderer(TocMixin, Renderer):
            pass

        toc = TocRenderer()
        md = mistune.Markdown(renderer=toc)

        # required in this order
        toc.reset_toc()          # initial the status
        md.parse(text)           # parse for headers
        toc.render_toc(level=3)  # render TOC HTML
    """

    def reset_toc(self):
        self.toc_tree = []
        self.toc_count = 0

    def header(self, text, level, raw=None):
        rv = '<h%d id="toc-%d">%s</h%d>\n' % (
            level, self.toc_count, text, level
        )
        self.toc_tree.append((self.toc_count, text, level, raw))
        self.toc_count += 1
        return rv

    def render_toc(self, level=3):
        """Render TOC to HTML.

        :param level: render toc to the given level
        """
        return ''.join(self._iter_toc(level))

    def _iter_toc(self, level):
        first_level = -1
        last_level = -1

        yield '<ul id="table-of-content">\n'

        for toc in self.toc_tree:
            index, text, l, raw = toc

            if l > level:
                # ignore this level
                continue

            if first_level == -1:
                # based on first level
                first_level = l
                last_level = l
                yield '<li><a href="#toc-%d">%s</a>' % (index, text)
            elif last_level == l:
                yield '</li>\n<li><a href="#toc-%d">%s</a>' % (index, text)
            elif last_level == l - 1:
                last_level = l
                yield '<ul>\n<li><a href="#toc-%d">%s</a>' % (index, text)
            elif last_level > l:
                # close indention
                yield '</li>'
                while last_level > l:
                    yield '</ul>\n</li>\n'
                    last_level -= 1
                yield '<li><a href="#toc-%d">%s</a>' % (index, text)

        # close tags
        yield '</li>\n'
        while last_level > first_level:
            yield '</ul>\n</li>\n'
            last_level -= 1

        yield '</ul>\n'



# Generated by Django 4.1.3 on 2022-12-15 05:02

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('inbox', '0001_initial'),
    ]

    operations = [
        migrations.AlterField(
            model_name='message',
            name='cc',
            field=models.TextField(default=None, null=True),
        ),
    ]



__version__ = "1.2"

GOOGLE_AUTH_URI = 'https://accounts.google.com/o/oauth2/auth'
GOOGLE_REVOKE_URI = 'https://accounts.google.com/o/oauth2/revoke'
GOOGLE_TOKEN_URI = 'https://accounts.google.com/o/oauth2/token'




from django.db import models
from django.conf import settings


from account.models import User, Member, Group, MemberFeed
from medialib.models import MediaItem, MediaLibrary
from location.models import GeoIP, Address

from rest import helpers as rest_helpers
from rest.middleware import get_request
from rest import RemoteEvents
from rest.models import RestModel, MetaDataModel, MetaDataBase


class Product(models.Model, RestModel, MetaDataModel):
    class RestMeta:
        GRAPHS = {
            "base": {
                "graphs": {
                    "current":"default",
                    "beta":"default",
                },
                "exclude":["library"]
            },
            "default": {
                "graphs": {
                    "current":"default",
                    "beta":"default",
                },
                "exclude":["library"]
            }
        }
    oid = models.CharField(max_length=250, db_index=True, blank=True, default="")
    created = models.DateTimeField(auto_now_add=True, editable=False)
    modified = models.DateTimeField(auto_now=True)

    owner = models.ForeignKey(Member, related_name="+", on_delete=models.CASCADE)
    group = models.ForeignKey(Group, related_name="+", blank=True, null=True, default=None, on_delete=models.CASCADE)

    archived = models.BooleanField(default=False, blank=True)

    name = models.CharField(max_length=200)
    description = models.TextField(blank=True, null=True, default=None)
    kind = models.CharField(max_length=64, blank=True, null=True, default=None)
    is_public = models.BooleanField(default=False, blank=True)

    current = models.ForeignKey("Release", related_name="+", blank=True, null=True, default=None, on_delete=models.CASCADE)
    beta = models.ForeignKey("Release", related_name="+", blank=True, null=True, default=None, on_delete=models.CASCADE)

    library = models.ForeignKey(MediaLibrary, blank=True, null=True, default=None, on_delete=models.CASCADE)

    def getReleases(self):
        return self.releases.all().order_by("-created")[:10]

    def getLibrary(self, commit=True):
        if not self.library:  # object is being created, thus no primary key field yet
            library = MediaLibrary(owner=self.owner, name="Releases")
            library.save()
            self.library = library
            if commit:
                self.save()
        return self.library

    def save(self, *args, **kwargs):
        self.getLibrary(False)
        super(Product, self).save(*args, **kwargs)

    def __unicode__(self):
        return "{0} - {1}".format(self.name, self.kind)


class ProductMetaData(MetaDataBase):
    parent = models.ForeignKey(Product, related_name="properties", on_delete=models.CASCADE)


class Release(models.Model, RestModel, MetaDataModel):
    class RestMeta:
        GRAPHS = {
            "default": {
                "exclude":["media"],
                "extra":["download_url"]
            }
        }

        POST_SAVE_FIELDS = ["media"]

    created = models.DateTimeField(auto_now_add=True, editable=False)
    modified = models.DateTimeField(auto_now=True)
    owner = models.ForeignKey(Member, related_name="+", on_delete=models.CASCADE)
    product = models.ForeignKey(Product, related_name="releases", on_delete=models.CASCADE)

    notes = models.TextField(blank=True, null=True, default=None)
    version_num = models.IntegerField(db_index=True, default=1)
    version_str = models.CharField(max_length=32, blank=True, null=True)

    media = models.ForeignKey(MediaItem, on_delete=models.PROTECT, blank=True, null=True, default=None)

    def set_product(self, value):
        if isinstance(value, str):
            self.product = Product.objects.filter(oid=value).last()
        elif isinstance(value, int):
            self.product = Product.objects.filter(pk=value).last()
        else:
            self.product = value

    def makeCurrent(self):
        self.product.current = self
        self.product.save()

    def makeBeta(self):
        self.product.beta = self
        self.product.save()

    def download_url(self):
        if self.media:
            return self.media.url()
        return None

    def on_rest_pre_save(self, request, **kwargs):
        self.owner = self.product.owner
        if self.version_num == 1:
            lr = Release.objects.filter(product=self.product).last()
            if lr is not None:
                self.version_num = lr.version_num + 1

    def on_rest_saved(self, request, is_new=False):
        if is_new:
            self.product.current = self
            self.product.save()



from django.contrib import admin
from . import models as inbox

admin.site.register(inbox.Mailbox)
admin.site.register(inbox.Message)
admin.site.register(inbox.Attachment)
admin.site.register(inbox.Bounce)
admin.site.register(inbox.Complaint)
